{{
xmodem loader for multiple propellers
Sets up a simple link so data can be downloaded and then passed out to slave propellers
Uses the xmodem protocol for simple error checking
Send commands via lines of text
}}
CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  _baudRateSpeed = 115200 ' don't use Teraterm at 115200 and possibly even 38400. Use Shamcom/hyperterminal
  _receiverPin = 31
  _transmitterPin = 30


  _RTC_DAT = -1 ' -1 ifnot installed.
  _RTC_CLK = -1 ' -1 ifnot installed.

  _TV_Enabled = true ' TV display on
  _TV_Pin = 16 ' group of 3 pins starting at pin 16

  Quote = 34
  Horizontal_Tab = 9
  Line_Feed = 10
  Carriage_Return =13

'prop loader constants
  #1, ErrorConnect, ErrorVersion, ErrorChecksum, ErrorProgram, ErrorVerify
  #0, Shutdown, LoadRun, ProgramShutdown, ProgramRun  

OBJ
  str:    "ASCII0_STREngine.spin"
  delay:  "Timing"                     ' for millisecond delays
  sio:    "pcFullDuplexSerial2FC"
  text:   "tv_text"    


VAR
   byte CommandLine[80] ' complete command line as typed in
   byte LineOfText1[80] ' general purpose string buffer
   'byte LineOfText2[80] ' general purpose string buffer
   'byte LineOfText3[80] ' general purpose string buffer
   'byte Filename[14]    ' filename
   word i

   long P31, P30, LFSR, Ver, Echo
   long CommCounter,Filesize,FirstPacket,Pingroup   
          

PUB shell
  dira := %00000000_00000000_00000000_00000000
  outa := %00000000_00000000_00000000_00000000

  text.start(_TV_Pin)                                 ' start the TV driver
  text.clearscreen
  sio.AddPort(0, _receiverPin, _transmitterPin, -1, -1, 0, 0, _baudRateSpeed)
  PrintstringCR(string("Multiprop Loader"))  
  sio.start             ' start the com port  note can't print anything to screen till sio.start unless only do to vga screen
  sio.rxflush(0)
  printstring(string("Baud rate = "))
  printdecimal(_baudratespeed)              ' print the baud rate
  crlf
  'delay.pause1ms(1000) ' so something comes up on the display
  'Printstringcr(string("Command = XM 0")) ' command, pingroup to send on (0,4,8 etc
  repeat  ' repeat until all files sent, then will overwrite with a transfer to the master prop anyway
    PrintstringCR(string("Wait for command"))
    result := shellLine ' ignore wakeup XM
    str.copy(result,@commandline)' save but discard
    crlf
    result := shellLine ' SIZE=nnnnn where always 5 bytes for size. Decimal
    str.copy(result,@commandline)
    crlf
    str.mid(@commandline,@lineoftext1,6,5)
    Filesize := str.decimaltointeger(@lineoftext1)
    'text.dec(size)
    result := shellLine ' PINGROUP=nnnnn where always 5 bytes for size. Decimal
    str.copy(result,@commandline)
    str.mid(@commandline,@lineoftext1,10,5)
    pingroup := str.decimaltointeger(@lineoftext1)
    crlf
    'text.dec(pingroup)
    FirstPacket := true ' flag for sending packets below
    XmodemReceive
    CommunicateFinish
    crlf



PRI shellLine
  repeat
      result:=0
      repeat
        if sio.rxavail(0)
          result :=sio.rx(0)  
      until result <>0           ' until something comes in 
    
    'if result==200
    '  result :=8                ' backspace keycode is 200 but convert to 8
    PrintChar(result)             ' print character
    'text.dec(result)            ' debugging
    str.buildString(result)     ' build an input string
  while((result <> 13) and (result <> 10) and (result <> 0))
  return str.trimString(str.builtstring(true)) ' reset the string
         

PUB PrintStringCR(pstring)' sends pstring to local display
     PrintString(pstring)                               ' print the string  
     crlf

PUB PrintString(pstring)' sends pstring to local display
    text.str(pstring)

PUB PrintChar(c) ' sends c to vga and serial port
    text.out(c)
    
PUB CRLF
    PrintChar(13)                                        ' send to vga
    PrintChar(10)                                        ' send to vga

PUB PrintDecimal(number)                                ' print the decimal value, useful for debuggin
     PrintString(str.integerToDecimal(number, 10))   

VAR
   byte XmodemBuffer[132]
   byte validpacket
   word PacketNumber
   byte ErrorCounter
   byte OnesComplement
   word Checksum
   word Timeout

PRI XmodemReceive
    Text.str(string("Send NAK and waiting for reply")) ' only print on local display, not to serial port
    i:=0                                                ' set counter to 0
    sio.rxflush(0)                                      ' clear the receive buffer
     repeat while (i < 15 and sio.rxavail(0)==0)        ' send 15 NAKs don't make the delay too short
       Text.out(".")                                   ' print a . on vga display
       sio.tx(0,21)                                     ' send a NAK to start things off
       delay.pause1ms(1000)                             ' wait 1 second 
       i++                                              ' add 1 to i
    if sio.rxavail(0)
        Text.str(string(13,10,"Starting file transfer",13,10))
       delay.pause1ms(100)                              ' short delay makes "1st packet errors" less likely                        
       errorcounter:=0                                  ' reset error counter
       packetnumber:=1                                   ' packet number 1
      repeat
         validpacket :=255                              ' equals true, set to 0 if any errors further down
         timeout:=0
         xmodembuffer[0] := 0                           ' definitely clear for startup                        
         repeat while timeout < 100                      ' timeout if no byte arrives must be <255
           if sio.rxavail(0)                              ' is there a byte?
             XmodemBuffer[0]:=sio.rx(0)                         ' read in one byte
             timeout:=65000                               ' finish the timeout
           else
             delay.pause1ms(1)                          ' wait 1 ms
             timeout++                                  ' add 1 to timeout
        if XmodemBuffer[0] <> 4                         ' finish byte
          if XmodemBuffer[0] <> 1                       ' not 1 so packet corrupt or timed out (0)
            validpacket:=0                              ' invalid packet
            errorcounter++                              ' add one to error counter
            Text.out("x")                              ' a common error on 1st packet so print something short
            Text.dec(XmodemBuffer[0])                  ' print out the first item in the buffer
            sendNAK                                     ' send a NAK/flush buffer and print

          else                                          ' 1= a valid start byte for a packet
            i:=1
            timeout:=0                                  ' if times out then packet will be corrupt
            repeat while i<132                         ' includes a timeout if half packet missing
              if sio.rxavail(0)                         ' if a byte is available
                XmodemBuffer[i]:=sio.rx(0)                    ' read it in
                i++                                       ' and increment
              else
                delay.pause1ms(10)                        ' otherwise, pause a little
                timeout++                                 ' and keep track of the pauses
                if timeout>100                            ' and give up after 1 second total
                  i:=133                                 ' force end if there is a timeout
                  errorcounter++                        ' and add one to error counter
                  Text.out("t")

            'repeat i from 0 to 131                      ' debugging
             'Text.hex(xmodembuffer[i],2)
             'Text.out(",")
                    
            packetnumber &= 255                         ' bitwise AND with 255, quicker than repeatedly subtracting 256      
            onescomplement:=255-packetnumber            ' calculate the ones complement
            if packetnumber <> XmodemBuffer[1]
              validpacket:=0                            ' not a valid packet
              errorcounter++
              Text.out("p")
            if XmodemBuffer[2] <> onescomplement            ' ones complement not correct
              validpacket:=0
              errorcounter++
              Text.out("o")                             
            checksum:=0                                 ' calculate the checksum
            repeat i from 3 to 130
               checksum:=checksum+XmodemBuffer[i]

            checksum &= 255                             ' bitwise AND with 255, same as repeatedly subtracting 255
            'repeat while checksum>255                   ' round to under 256
                'checksum:=checksum-256
            if checksum <> XmodemBuffer[131]                   ' checksum not valid
              validpacket:=0
              errorcounter++
              Text.out("c")
            if validpacket == 255                         ' packet is valid so save it
              DumpData
              packetnumber++                             ' increment packet number
              sio.tx(0,6)                                ' send a ACK and wait for the next packet
              Text.out(".")                             ' print a . on the screen for each packet
            else
              sendNAK
              Text.out("N")
      until errorcounter>9 or xmodembuffer[0]==4 ' too many errors, or 04=the finish byte
       sio.tx(0,6)
      if errorcounter==0
        Text.str(string(13,10,"No XM errors"))
      if errorcounter>0 and errorcounter<3
        Text.str(string(13,10,"File OK, some corrected errors"))
      if errorcounter>2
        repeat i from 1 to 10
          Text.str(string("Cancel "))
          sio.tx(0,18)                                  ' Send Cancel 10 times



PRI SendNAK
    sio.rxflush(0)
    sio.tx(0,21)                               ' send a NAK as the packet was corrupted
    errorcounter++
    Text.str(string(" NAK "))       ' increment the error counter
    delay.pause1ms(500)  


PUB DumpData  
    'repeat i from 3 to 130 ' discard first 3 bytes and last byte 
    '  Text.hex(XmodemBuffer[i],2)
    CommunicateSendPacket

' *************** prop loader code *******************
PRI CommunicateStart(Version, Command) | ByteCount,PinRESn, PinP31, PinP30
  pinRESn:= Pingroup
  pinP31 := Pingroup+1
  pinP30 := Pingroup+2
  P31 := PinP31                                         'set P31 and P30 
  P30 := PinP30
  outa[PinRESn] := 0                                    ' RESn low
  dira[PinRESn] := 1                                    ' and set as output
  outa[PinP31] := 1                                     'P31 high (our TX)     
  dira[PinP31] := 1                                     ' and set as output
  dira[PinP30] := 0                                     'P30 input (our RX)   
  outa[PinRESn] := 1                                    'RESn high   
  waitcnt(clkfreq / 10 + cnt)                           ' wait 100ms

  CommCounter := 0          ' global counter
  BitsOut(%01, 2)            'output calibration pulses       
  LFSR := "P"                 'send LFSR pattern          
  repeat 250
    BitsOut(IterateLFSR, 1)
  repeat 250                     'receive and verify LFSR pattern    
    if WaitBit(1) <> IterateLFSR
      text.str(string("Error Connect"))
  repeat 8                  'receive chip version      
    Ver := WaitBit(1) << 7 + Ver >> 1
  text.str(string("Version "))
  text.dec(Ver)
  crlf
  if Ver <> Version                  'if version mismatch, shutdown and abort    
    BitsOut(Shutdown, 32)
    text.str(string(" Comm Error Version "))
  BitsOut(Command, 32)                                    'send command                              
  if Command                                              'handle command details     
    'ByteCount := byte[CodePtr][8] | byte[CodePtr][9] << 8   ' length of file
    ByteCount := Filesize                            ' number of bytes to send
    BitsOut(ByteCount >> 2, 32)

PRI CommunicateSendPacket | n ' send 128 bytes or quit if more than number sent
   if firstpacket == true ' so no delays waiting for xmodem
     CommunicateStart(1, LoadRun) ' start up the loader
     firstpacket := false

   n := 128
   if Filesize - CommCounter < 128
     n := Filesize - CommCounter 
   repeat i from 1 to n                                   'send bytes    
      BitsOut(XmodemBuffer[i+2],8)                        ' read byte from xmodem buffer
      'text.dec(XmodemBuffer[i+2])
      'text.out(32) 
   CommCounter += 128
   
PRI CommunicateFinish| PinRESn, PinP31, PinP30
  pinRESn:= Pingroup
  pinP31 := Pingroup+1
  pinP30 := Pingroup+2
  if WaitBit(25)                                          'allow 250ms for positive checksum response    
    text.str(string(" Comm Error "))
  dira[PinRESn] := 0                                    ' reset pin floating
  dira[PinP31] := 0                                     'P31 float  

  
  


PRI IterateLFSR : Bit
  'get return bit
  Bit := LFSR & 1
  'iterate LFSR (8-bit, $B2 taps)
  LFSR := LFSR << 1 | (LFSR >> 7 ^ LFSR >> 5 ^ LFSR >> 4 ^ LFSR >> 1) & 1

PRI WaitBit(Hundredths) : Bit | PriorEcho
  repeat Hundredths
    'output 1t pulse                        
    BitsOut(1, 1)
    'sample bit and echo
    Bit := ina[P30]
    PriorEcho := Echo
    'output 2t pulse
    BitsOut(0, 1)
    'if echo was low, got bit                                      
    if not PriorEcho
      return
    'wait 10ms
    waitcnt(clkfreq / 100 + cnt)
  'timeout, abort
  text.str(string(" Comm Timeout "))

  
PRI BitsOut(Value, Bits)
  repeat Bits
    if Value & 1
      'output '1' (1t pulse)
      outa[P31] := 0                        
      Echo := ina[P30]
      outa[P31] := 1
    else
      'output '0' (2t pulse)
      outa[P31] := 0
      outa[P31] := 0
      Echo := ina[P30]
      Echo := ina[P30]
      outa[P31] := 1
    Value >>= 1
                   
        
{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}